"""
Authors: Simrat Hanspal and Rudy Venguswamy
CeleryStalk Framework v1.1.1
Contact: simrath@vmware.com, rudy@ischool.berkeley.edu

"""

from __future__ import absolute_import, unicode_literals
# from celery.task.control import inspect

import os
import sys
import pdb
import time
import emoji
import traceback
import subprocess
from itertools import chain
from argparse import ArgumentParser


class Runner:
    def __init__(self, task_filepath, ip_loc, op_loc):
        """
       Init Runner for Celery & Tasks File
       :param task_filepath: path for parallelizing task
       :param runner_filepath: this file
       :param ip_loc: CSV file location
       :param op_loc: where you want the CSV file processed to end up
       """
        self.ip_loc = ip_loc
        self.op_loc = op_loc
        self.task_filepath = task_filepath
        
    def setup(self):
        """
        Starts the Redis Server and Celery Workers
        :return: No return
        """


        print("Starting setup ...")

        # ==== LOAD TASK MODULE ==== #
        if self.task_filepath not in sys.path:
            sys.path.append(self.task_filepath)
        filename = self.task_filepath.split("/")[-1] if "/" in self.task_filepath else self.task_filepath
        module_name = filename.replace(".py", "")
        self.module = __import__(module_name)
        print("Task module loaded")

        # ==== START REDIS SERVER ==== #
        cmd = "nohup redis-server&"
        returned_value = subprocess.call(cmd, shell=True)
        print("\nRedis server started with status =", returned_value)

        # ==== START REDIS SERVER ==== #
        cmd = "celery worker -A parallelised_pandas_apply -E -n CeleryStalkWorker -l INFO --detach "
        returned_value = subprocess.call(cmd, shell=True)
        print("Celery worker threads started with status =", returned_value)

        time.sleep(0.1)
        print(emoji.emojize(":triangular_flag_on_post: SETUP COMPLETE \n", use_aliases=True))

    def load_input_files(self):
        """
        Runs the master_run method from the task file (parallelise_pandas_apply.py provided with package)
        :return: no return
        """

        self.module.master_run(self.op_loc, self.ip_loc)
        return

    def execute(self):
        """
        calls load_input_files
        :return: no return
        """
        start = time.time()
        self.load_input_files()
        print(emoji.emojize(":triangular_flag_on_post: EXECUTION COMPLETE \n", use_aliases=True))
        print('TIME TAKEN ::', time.time()-start)
        
    def teardown(self):
        """
        kills celery workers and redis server at the end of the task process
        :return: no return
        """

        cmd = "ps aux | grep python | grep 'CeleryStalkWorker' | grep -v 'grep'| awk '{print $2}' | xargs kill"
        returned_value = subprocess.call(cmd, shell=True)
        print("Celery workers shutdown with status =", returned_value)

        ps = subprocess.Popen("ps aux | grep python | grep 'CeleryStalkWorker' | grep -v 'grep'| awk '{print $2}'", shell=True, stdout=subprocess.PIPE)
        process_status = ps.stdout.read().decode("utf-8")
        while process_status != "":
            print("Celery processes are shutting down. Sleeping for 1 sec ... zzZ")
            time.sleep(1)
            ps = subprocess.Popen("ps aux | grep python | grep 'CeleryStalkWorker' | grep -v 'grep'| awk '{print $2}'", shell=True, stdout=subprocess.PIPE)
            process_status = ps.stdout.read().decode("utf-8")

        cmd = "pkill -f redis-server"
        returned_value = subprocess.call(cmd, shell=True)
        print("Redis Server shutdown with status =", returned_value)

        if os.path.exists("nohup.out"):
            os.remove("nohup.out")
        if os.path.exists("dump.rdb"):
            os.remove("dump.rdb")

        sys.path.remove(self.task_filepath)
        time.sleep(0.1)
        print(emoji.emojize(":triangular_flag_on_post: TEARDOWN COMPLETE \n", use_aliases=True))
        
    def _process(self):
        """
        calls setup function
        :return: no return
        """
        # -- SETUP --#
        try:
            self.setup()
        except Exception as e:
            print(emoji.emojize(":x: Error encountered in SETUP block.\nError = %(e)s.\nTraceback = %(t)s" % {"e": str(e),
                                                                                                              "t": str(traceback.format_exc())}, use_aliases=True))
            return

        # -- EXECUTE -- #
        try:
            self.execute()
        except Exception as e:
            print(emoji.emojize(":x: Error encountered in EXECUTE block.\nError = %(e)s.\nTraceback = %(t)s" % {"e": str(e),
                                                                                                                "t": str(traceback.format_exc())}, use_aliases=True))
            return

    def main(self):
        """
        iniates the process and teardown process
        :return: no return
        """
        self._process()
        self.teardown()
        print(emoji.emojize("Done :thumbs_up: \n", use_aliases=True))


if __name__ == "__main__":
    """
    provides argument list for command line based execution
    :return: no return
    """
    parser = ArgumentParser(description='Parallel processing framework built on top of Celery.')
    parser.add_argument('--task_file', required=True, dest="task_file",
                        help='task file containing function to be parallelised.')
    parser.add_argument('--ip_loc',  required=True, dest='ip_loc',
                        help='input folder location')
    parser.add_argument('--op_loc', required=True, dest='op_loc',
                        help='output folder location')

    args = parser.parse_args()
    r = Runner(args.task_file, args.ip_loc, args.op_loc)
    r.main()



#!/usr/bin/env python
# coding: utf-8
"""
Authors: Simrat Hanspal and Rudy Venguswamy
CeleryStalk Framework v1.0.1
Contact: simrath@vmware.com, rudy@ischool.berkeley.edu

"""
from __future__ import absolute_import, unicode_literals
from celery import group
from celery import Celery
from celery.utils.log import get_task_logger

import gc
import os
import csv
import pdb
import time
import dill as dill
import pandas as pd
import multiprocessing
import jsonpickle as jp

from celery.task.control import inspect
"""
Initiates the celery app identifying process
"""

app = Celery('CeleryStalk',
             broker='redis://localhost:6379/0',
             backend='redis',
             include=[])


#logs celery tasks
logger = get_task_logger(__name__)

# Optional configuration, see the application user guide for celery.
app.conf.update(
    result_expires=3600,
)


@app.task
def readtodf(jsonfn, workdirectory, filenamed):
    """
    takes a chunk of the CSV, which is a piece of the original CSV, split, reads it to a dataframe, applies the function, and returns it to CSV
    :param jsonfn: json version of a function
    :param workdirectory: where splitCSV and processedCSV (intermediate CSV files) get stored
    :param filenamed: name of CSV to be processed
    """
    url = workdirectory + '/SplitCSV/' + filenamed
    df = pd.read_csv(url)
    # unpack the lambda function from json to pickle to function
    picklefn = jp.decode(jsonfn)
    func = dill.loads(picklefn)
    df = df.apply(func, axis=1)
    os.chdir(workdirectory + 'SplitCSVProcessed/')
    print('Function Applied. Now Saving Results')
    df.to_csv(filenamed + 'processed.csv', encoding='utf-8')
    return 'Done Processing CSVs. Merging them...'


#takes all processed CSV files and joins them together
@app.task
def concatdf(workdirectory):
    """
    Takes all the files in a directory and concatenates them into one CSV
    :param workdirectory: directory where the folder SplitCSVProcessed exists, used for the merge into a final large, processed csv
    """
    os.chdir(workdirectory)
    fout = open("out.csv", 'ab')
    # first file:
    first = True
    os.chdir(workdirectory + '/SplitCSVProcessed')
    print('JOINING FILES')
    for f in os.listdir(workdirectory + '/SplitCSVProcessed'):
        with open(f, 'rb') as f:
            for line in f:
                fout.write(line)
    fout.close()

    return 'Done with Function'


def splitcsv(workdirectory, file, num_cores=multiprocessing.cpu_count()):
    """
    splits a large CSV into smaller CSVs to allow for parallel processing
    :param workdirectory: file location for intermediate CSV folders (splitCSV, SplitCSVprocessed) to be created
    :param file: file to be processed via parallelization
    :param num_cores: number of cpus that can be used for parallelizing pandas apply.
    """
    os.chdir(workdirectory)
    with open(file, 'r') as f:
        csvfile = f.readlines()
    print('NUM_CORES USED FOR PARALLELIZING: ', num_cores)
    csvlen = len(csvfile)
    linesPerFile = int(csvlen / num_cores) + 1
    filename = 1
    os.chdir(workdirectory + '/SplitCSV')
    for i in range(0, csvlen, linesPerFile):
        with open(str(filename) + '.csv', 'w+') as f:
            f.writelines(csvfile[i:i + linesPerFile])
        filename += 1
    return 'CSV Split into Chunks for Processing...'


def examplefn(vec):
    """
    example function that will be applied across CSV chunks
    :param vec: a row of a data frame, used inside a pandas.apply(function, axis = 1)
    """
    vec[0] = vec[0] + 1000
    return vec


def c_apply(workdirectory, csvdirectory, lambdafn=examplefn):
    """
    creates a directory for Split CSVs, processes SPLIT CSVs and initiates the entire parallel process
    :param workdirectory: file location for intermediate CSV folders (splitCSV, SplitCSVprocessed) to be created
    :param csvdirectory: location of the file to be processed in parallel
    :param lambdafn: function to be applied across a pandas dataframe 
    """
    # turns function into celery passable object
    picklefn = dill.dumps(lambdafn)
    jsonfn = jp.encode(picklefn)

    if not os.path.exists(workdirectory + '/SplitCSV'):
        os.makedirs(workdirectory + '/SplitCSV')

    if not os.path.exists(workdirectory + '/SplitCSVProcessed'):
        os.makedirs(workdirectory + '/SplitCSVProcessed')

    print(splitcsv(workdirectory, csvdirectory))
    print('Reading and Modifying Files')

    filelist = []
    for filename in os.listdir(workdirectory + '/SplitCSV'):
        filelist.append(filename)
    i = inspect()
    print('TASK REGISTER',i.registered_tasks())
    #The following 3 lines execute the parallel process: run the readtodf method in parallel, and then after all are executed, concat the CSVS into one big CSV again
    c3 = (group(readtodf.si(jsonfn, workdirectory, filename) for filename in filelist) | concatdf.si(workdirectory))

    res = c3()
    value = res.get()
    del res
    gc.collect()
    return value
    

def master_run(workdirectory, csvdirectory):
    """
    Redundant method that calls c_apply but allows for a generic framework that eliminates the need to modify parallelise.py if c_apply is modified.
    :param workdirectory: file location for intermediate CSV folders (splitCSV, SplitCSVprocessed) to be created
    :param csvdirectory: location of the file to be processed in parallel
    """
    c_apply(workdirectory, csvdirectory, examplefn)
    return



from setuptools import setup



setup(
    name='CeleryStalk',
    version='1.3.6',
    author='Rudy Venguswamy',
    author_email='rvenguswamy@vmware.com',
    packages=['CeleryStalk'],
    scripts=[],
    url='http://pypi.python.org/pypi/CeleryStalk/',
    license='LICENSE.txt',
    description='Change Large CSVs Faster with Easy Parallel Processing',
    long_description="README.txt",
    install_requires=[
        "celery >= 4.2.0",
        "redis >= 2.10.5",
    ],
)